;; FM Radio (HAM Radio)
;;
;; Copyright 2019 - By Michael Kohn
;; http://www.mikekohn.net/
;; mike@mikekohn.net
;;
;; Using a SparkFun FM Tuner Basic Breakout - Si4703 (BOB-11083),
;; create an FM radio inside a chunk of ham.  An LM386 audio is
;; used to interface mono to a speaker and a potentiometer hooked
;; up to the ADC of the MSP430 is used for tuning.

.include "msp430x2xx.inc"

; 80.0 ms = 1 interrupts

RAM equ 0x0200
CHANNEL equ RAM
CHANNEL_LAST equ RAM+2
DEVICE_ID equ RAM+16
CHIP_ID equ RAM+18
CURRENT_FREQ equ RAM+20
DEBUG_0 equ RAM+32
DEBUG_1 equ RAM+34
DEBUG_2 equ RAM+36

FM_SEN  equ 0x20
FM_CLK  equ 0x10
FM_DATA equ 0x08
FM_RST  equ 0x04

;  r4 =
;  r5 =
;  r6 =
;  r7 =
;  r8 =
;  r9 =
; r10 = Temp in read_adc, counter in 3wire code
; r11 = Temp in read_adc
; r12 =
; r13 =
; r14 =
; r15 = temp

  .org 0xf800
start:
  ;; Turn off watchdog
  mov.w #(WDTPW|WDTHOLD), &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Setup stack pointer
  mov.w #0x0280, SP

  ;; Set MCLK to 1.4 MHz with DCO
  mov.b #DCO_3, &DCOCTL
  mov.b #RSEL_8, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Setup output pins
  ;; P1.3 = 3 wire DATA
  ;; P1.4 = 3 wire CLK
  ;; P1.5 = 3 wire /SEN (chip select)
  ;; P1.6 = ADC for tuner potentiometer
  mov.b #FM_SEN|FM_CLK|FM_DATA|FM_RST, &P1DIR
  mov.b #FM_SEN|FM_RST, &P1OUT
  mov.b #0x00, &P1SEL
  mov.b #0x00, &P1SEL2

  ;; Setup Timer
  ;mov.w #40000, &TACCR0
  ;mov.w #(TASSEL_2|ID_2|MC_1), &TACTL ; SMCLK, DIV1, COUNT to TACCR0
  ;mov.w #CCIE, &TACCTL0
  ;mov.w #0, &TACCTL1

  ;; Setup A/D
  mov.w #ADC10ON|ADC10SHT_3, &ADC10CTL0
  mov.w #INCH_6|ADC10SSEL_0, &ADC10CTL1
  mov.w #0x40, &ADC10AE0

  ;; Force a reset on the FM radio device.
  call #delay
  bic.b #FM_RST, &P1OUT
  bic.b #FM_SEN, &P1OUT
  bis.b #FM_DATA, &P1OUT
  call #delay_1ms
  bis.b #FM_RST, &P1OUT
  bis.b #FM_SEN, &P1OUT
  bic.b #FM_DATA, &P1OUT
  call #delay_1ms

  ;; Turn on interrupts
  eint

  ;; Get Device ID
  mov.b #0, r15
  call #read_3wire
  mov.w r15, &DEVICE_ID

  ;; Turn on crystal oscillator.  Read register first and set bit 15
  ;; then rewrite.
  mov.b #7, r15
  call #read_3wire
  mov.w r15, r14
  mov.w r15, &DEBUG_2
  mov.b #7, r15
  bis.w #0x8000, r14
  call #write_3wire

  ;; Wait 840ms for oscillator to start up.
  call #delay
  call #delay
  call #delay
  call #delay
  call #delay
  call #delay

  ;; Powerup.
  mov.b #2, r15
  mov.w #0xe001, r14
  call #write_3wire

  call #delay

  ;; Get Chip ID.  Bit 6 should be 1 on powerup.
  mov.b #1, r15
  call #read_3wire
  mov.w r15, &CHIP_ID

  ;; Set to RSSI=0, band=USA, spacing=200kHz (0.2 MHz), full volume. 
  mov.b #5, r15
  mov.w #0x000f, r14
  call #write_3wire

  mov.w #1024, &CHANNEL_LAST

main:
  call #read_adc
  call #delay
  jmp main

;; @1.4MHz this is about 140ms
delay:
  mov.w #0, r15
delay_loop:
  dec.w r15
  jnz delay_loop
ret

delay_1ms:
  mov.w #467, r15
delay_loop_1ms:
  dec.w r15
  jnz delay_loop_1ms
ret

;; read_adc()
read_adc:
  mov.w #INCH_6|ADC10SSEL_0|ADC10DIV_7, &ADC10CTL1
  bic.w #ADC10IFG, &ADC10CTL0
  bis.w #ENC|ADC10SC, &ADC10CTL0
read_adc_busy:
  bit.w #ADC10BUSY, &ADC10CTL1
  jnz read_adc_busy
  mov.w &ADC10MEM, r10
  bic.w #ENC, &ADC10CTL0
  rra.w r10
  rra.w r10
  mov.w r10, r11
  sub.w &CHANNEL_LAST, r11
  add.w #4, r11
  cmp.w #8, r11
  jlo read_adc_dont_update_channel
  mov.w r10, &CHANNEL_LAST
  rra.w r10
  rra.w r10
  rra.w r10
  rra.w r10
  mov.b r10, &CHANNEL

  mov.w #0x000f, r11
  sub.w r10, r11
  rla.w r11
  add.w #channels, r11
  mov.w @r11, r11
  mov.w r11, &DEBUG_0

  ;; Set channel
  mov.w r11, r15
  call #set_channel

read_adc_dont_update_channel:
  ret

;; read_3wire(address.b=r15) -> r15
read_3wire:
  bis.b #0x20, r15
  bic.b #FM_SEN, &P1OUT
  call #send_address_3wire
  bic.b #FM_DATA, &P1DIR
  mov.w #0, r15
  mov.b #16, r10
read_3wire_loop:
  bis.b #FM_CLK, &P1OUT
  rla.w r15
  bit.b #FM_DATA, &P1IN
  jz read_3wire_0
  bis.w #1, r15
read_3wire_0:
  bic.b #FM_CLK, &P1OUT
  dec.b r10
  jnz read_3wire_loop
  call #reset_3wire
  ret

;; write_3wire(address.b=r15, data.w=r14)
write_3wire:
  bic.b #FM_SEN, &P1OUT
  call #send_address_3wire
  mov.b #16, r10
write_3wire_loop:
  bic.b #FM_DATA, &P1OUT
  bit.w #0x8000, r14
  jz write_3wire_loop_0
  bis.b #FM_DATA, &P1OUT
write_3wire_loop_0:
  bis.b #FM_CLK, &P1OUT
  rla.w r14
  bic.b #FM_CLK, &P1OUT
  dec.b r10
  jnz write_3wire_loop
  call #reset_3wire
  ret

;; send_address_3wire(address.b=r15)
;; Address is: 0 11[r/w]0 rrrr
send_address_3wire:
  bis.w #0x00c0, r15
  mov.b #9, r10
send_address_3wire_loop:
  bic.b #FM_DATA, &P1OUT
  bit.w #0x100, r15
  jz send_address_3wire_0
  bis.b #FM_DATA, &P1OUT
send_address_3wire_0:
  bis.b #FM_CLK, &P1OUT
  rla.w r15
  bic.b #FM_CLK, &P1OUT
  dec.b r10
  jnz send_address_3wire_loop
  bic.b #FM_DATA, &P1OUT
  ret

reset_3wire:
  bis.b #FM_SEN, &P1OUT
  bis.b #FM_CLK, &P1OUT
  bic.b #FM_CLK, &P1OUT
  bic.b #FM_DATA, &P1OUT
  bis.b #FM_DATA, &P1DIR
  ret

;; set_channel(channel=r15)
set_channel:
  mov.w r15, r4

  ;; Set channel ((Freq - 87.5) / 0.2.
  mov.b #3, r15
  mov.w r4, r14
  bis.w #0x8000, r14
  call #write_3wire

  ;; Wait for STC to go high.
set_channel_wait:
  mov.b #0xa, r15
  call #read_3wire
  mov.w r15, &DEBUG_1
  bit.w #0x4000, r15
  jz set_channel_wait

  ;; Clear STC by setting TUNE to 0.
  mov.b #3, r15
  mov.w r4, r14
  call #write_3wire

  ;; Get the current channel freq.
  mov.b #0xb, r15
  call #read_3wire
  mov.w r15, &CURRENT_FREQ
  ret

timer_interrupt:
  ;; Not used
  reti

;; Channels: ((Freq - 87.5) / 0.2.
channels:
  dw 0x10 ; 0) Public
  dw 0x1b ; 1) KWRH (Webster)
  dw 0x1f ; 2) The Bull
  dw 0x24 ; 3) KSHE
  dw 0x26 ; 4) Praise
  dw 0x2c ; 5) KNOU Top-40
  dw 0x2e ; 6) Religous (Crestwood)
  dw 0x38 ; 7) Talk
  dw 0x40 ; 8) The Beat
  dw 0x48 ; 9) Hip hop
  dw 0x4b ; 10) KEZK
  dw 0x4f ; 11) Klou
  dw 0x51 ; 12) Magic
  dw 0x57 ; 13) Alt
  dw 0x59 ; 14) The Answer
  dw 0x65 ; 15) Z107

.org 0xfff2
  dw timer_interrupt       ; Timer_A2 TACCR0, CCIFG
.org 0xfffe
  dw start                 ; Reset

